import copy
import logging

from kasa.deviceconfig import DeviceConfig
from kasa.protocols import IotProtocol
from kasa.transports.basetransport import BaseTransport

_LOGGER = logging.getLogger(__name__)


def get_realtime(obj, x, *args):
    return {
        "current": 0.268587,
        "voltage": 125.836131,
        "power": 33.495623,
        "total": 0.199000,
    }


def get_monthstat(obj, x, *args):
    if x["year"] < 2016:
        return {"month_list": []}

    return {
        "month_list": [
            {"year": 2016, "month": 11, "energy": 1.089000},
            {"year": 2016, "month": 12, "energy": 1.582000},
        ]
    }


def get_daystat(obj, x, *args):
    if x["year"] < 2016:
        return {"day_list": []}

    return {
        "day_list": [
            {"year": 2016, "month": 11, "day": 24, "energy": 0.026000},
            {"year": 2016, "month": 11, "day": 25, "energy": 0.109000},
        ]
    }


emeter_support = {
    "get_realtime": get_realtime,
    "get_monthstat": get_monthstat,
    "get_daystat": get_daystat,
}


def get_realtime_units(obj, x, *args):
    return {"power_mw": 10800}


def get_monthstat_units(obj, x, *args):
    if x["year"] < 2016:
        return {"month_list": []}

    return {
        "month_list": [
            {"year": 2016, "month": 11, "energy_wh": 32},
            {"year": 2016, "month": 12, "energy_wh": 16},
        ]
    }


def get_daystat_units(obj, x, *args):
    if x["year"] < 2016:
        return {"day_list": []}

    return {
        "day_list": [
            {"year": 2016, "month": 11, "day": 24, "energy_wh": 20},
            {"year": 2016, "month": 11, "day": 25, "energy_wh": 32},
        ]
    }


emeter_units_support = {
    "get_realtime": get_realtime_units,
    "get_monthstat": get_monthstat_units,
    "get_daystat": get_daystat_units,
}


emeter_commands = {
    "emeter": emeter_support,
    "smartlife.iot.common.emeter": emeter_units_support,
}


def error(msg="default msg"):
    return {"err_code": -1323, "msg": msg}


def success(res):
    if res:
        res.update({"err_code": 0})
    else:
        res = {"err_code": 0}
    return res


# plugs and bulbs use a different module for time information,
# so we define the contents here to avoid repeating ourselves
TIME_MODULE = {
    "get_time": {
        "year": 2017,
        "month": 1,
        "mday": 2,
        "hour": 3,
        "min": 4,
        "sec": 5,
    },
    "get_timezone": {
        "zone_str": "test",
        "dst_offset": -1,
        "index": 12,
        "tz_str": "test2",
    },
}

CLOUD_MODULE = {
    "get_info": {
        "username": "",
        "server": "devs.tplinkcloud.com",
        "binded": 0,
        "err_code": 0,
        "cld_connection": 0,
        "illegalType": -1,
        "stopConnect": -1,
        "tcspStatus": -1,
        "fwDlPage": "",
        "tcspInfo": "",
        "fwNotifyType": 0,
    }
}

SCHEDULE_MODULE = {
    "get_next_action": {
        "action": 1,
        "err_code": 0,
        "id": "0794F4729DB271627D1CF35A9A854030",
        "schd_time": 68927,
        "type": 2,
    },
    "get_rules": {
        "enable": 1,
        "err_code": 0,
        "rule_list": [
            {
                "eact": -1,
                "enable": 1,
                "id": "8AA75A50A8440B17941D192BD9E01FFA",
                "name": "name",
                "repeat": 1,
                "sact": 1,
                "smin": 1027,
                "soffset": 0,
                "stime_opt": 2,
                "wday": [1, 1, 1, 1, 1, 1, 1],
            },
        ],
        "version": 2,
    },
}

AMBIENT_MODULE = {
    "get_current_brt": {"value": 26, "err_code": 0},
    "get_config": {
        "devs": [
            {
                "hw_id": 0,
                "enable": 0,
                "dark_index": 1,
                "min_adc": 0,
                "max_adc": 2450,
                "level_array": [
                    {"name": "cloudy", "adc": 490, "value": 20},
                    {"name": "overcast", "adc": 294, "value": 12},
                    {"name": "dawn", "adc": 222, "value": 9},
                    {"name": "twilight", "adc": 222, "value": 9},
                    {"name": "total darkness", "adc": 111, "value": 4},
                    {"name": "custom", "adc": 2400, "value": 97},
                ],
            }
        ],
        "ver": "1.0",
        "err_code": 0,
    },
}


MOTION_MODULE = {
    "get_adc_value": {"value": 50, "err_code": 0},
    "get_config": {
        "enable": 0,
        "version": "1.0",
        "trigger_index": 2,
        "cold_time": 60000,
        "min_adc": 0,
        "max_adc": 4095,
        "array": [80, 50, 20, 0],
        "err_code": 0,
    },
}

LIGHT_DETAILS = {
    "color_rendering_index": 80,
    "err_code": 0,
    "incandescent_equivalent": 60,
    "lamp_beam_angle": 150,
    "max_lumens": 800,
    "max_voltage": 120,
    "min_voltage": 110,
    "wattage": 10,
}

DEFAULT_BEHAVIOR = {
    "err_code": 0,
    "hard_on": {"mode": "circadian"},
    "soft_on": {"mode": "last_status"},
}


class FakeIotProtocol(IotProtocol):
    def __init__(self, info, fixture_name=None, *, verbatim=False):
        super().__init__(
            transport=FakeIotTransport(info, fixture_name, verbatim=verbatim),
        )

    async def query(self, request, retry_count: int = 3):
        """Implement query here so tests can still patch IotProtocol.query."""
        resp_dict = await self._query(request, retry_count)
        return resp_dict


class FakeIotTransport(BaseTransport):
    def __init__(self, info, fixture_name=None, *, verbatim=False):
        super().__init__(config=DeviceConfig("127.0.0.123"))
        info = copy.deepcopy(info)
        self.discovery_data = info
        self.fixture_name = fixture_name
        self.writer = None
        self.reader = None
        self.verbatim = verbatim

        # When True verbatim will bypass any extra processing of missing
        # methods and is used to test the fixture creation itself.
        if verbatim:
            self.proto = copy.deepcopy(info)
        else:
            self.proto = self._build_fake_proto(info)

    @staticmethod
    def _build_fake_proto(info):
        """Create an internal protocol with extra data not in the fixture."""
        proto = copy.deepcopy(FakeIotTransport.baseproto)

        for target in info:
            if target != "discovery_result":
                for cmd in info[target]:
                    # Use setdefault in case the fixture has modules not yet
                    # part of the baseproto.
                    proto.setdefault(target, {})[cmd] = info[target][cmd]

        # if we have emeter support, we need to add the missing pieces
        for module in ["emeter", "smartlife.iot.common.emeter"]:
            if (
                module in info
                and "err_code" in info[module]
                and info[module]["err_code"] != 0
            ):
                proto[module] = info[module]
            else:
                for etype in ["get_realtime", "get_daystat", "get_monthstat"]:
                    if (
                        module in info and etype in info[module]
                    ):  # if the fixture has the data, use it
                        # print("got %s %s from fixture: %s" % (module, etype, info[module][etype]))
                        proto[module][etype] = info[module][etype]
                    else:  # otherwise fall back to the static one
                        dummy_data = emeter_commands[module][etype]
                        # print("got %s %s from dummy: %s" % (module, etype, dummy_data))
                        proto[module][etype] = dummy_data
        return proto

    @property
    def default_port(self) -> int:
        return 9999

    @property
    def credentials_hash(self) -> None:
        return None

    def set_alias(self, x, child_ids=None):
        if child_ids is None:
            child_ids = []
        _LOGGER.debug("Setting alias to %s, child_ids: %s", x["alias"], child_ids)
        if child_ids:
            for child in self.proto["system"]["get_sysinfo"]["children"]:
                if child["id"] in child_ids:
                    child["alias"] = x["alias"]
        else:
            self.proto["system"]["get_sysinfo"]["alias"] = x["alias"]

    def set_relay_state(self, x, child_ids=None):
        if child_ids is None:
            child_ids = []
        _LOGGER.debug("Setting relay state to %s", x["state"])

        _LOGGER.info("child_ids: %s", child_ids)
        if child_ids:
            for child in self.proto["system"]["get_sysinfo"]["children"]:
                if child["id"] in child_ids:
                    _LOGGER.info("Found %s, turning to %s", child, x["state"])
                    child["state"] = x["state"]
        else:
            self.proto["system"]["get_sysinfo"]["relay_state"] = x["state"]

    def set_led_off(self, x, *args):
        _LOGGER.debug("Setting led off to %s", x)
        self.proto["system"]["get_sysinfo"]["led_off"] = x["off"]

    def set_mac(self, x, *args):
        _LOGGER.debug("Setting mac to %s", x)
        self.proto["system"]["get_sysinfo"]["mac"] = x["mac"]

    def set_hs220_brightness(self, x, *args):
        _LOGGER.debug("Setting brightness to %s", x)
        self.proto["system"]["get_sysinfo"]["brightness"] = x["brightness"]

    def set_hs220_dimmer_transition(self, x, *args):
        _LOGGER.debug("Setting dimmer transition to %s", x)
        brightness = x["brightness"]
        if brightness == 0:
            self.proto["system"]["get_sysinfo"]["relay_state"] = 0
        else:
            self.proto["system"]["get_sysinfo"]["relay_state"] = 1
            self.proto["system"]["get_sysinfo"]["brightness"] = x["brightness"]

    def set_lighting_effect(self, effect, *args):
        _LOGGER.debug("Setting light effect to %s", effect)
        self.proto["system"]["get_sysinfo"]["lighting_effect_state"] = dict(effect)

    def transition_light_state(self, state_changes, *args):
        # Setting the light state on a device will turn off any active lighting effects.
        # Unless it's just the brightness in which case it will update the brightness for
        # the lighting effect
        if lighting_effect_state := self.proto["system"]["get_sysinfo"].get(
            "lighting_effect_state"
        ):
            if (
                "hue" in state_changes
                or "saturation" in state_changes
                or "color_temp" in state_changes
            ):
                lighting_effect_state["enable"] = 0
            elif (
                lighting_effect_state["enable"] == 1
                and state_changes.get("on_off") != 0
                and (brightness := state_changes.get("brightness"))
            ):
                lighting_effect_state["brightness"] = brightness
                return

        _LOGGER.debug("Setting light state to %s", state_changes)
        light_state = self.proto["system"]["get_sysinfo"]["light_state"]

        _LOGGER.debug("Current light state: %s", light_state)
        new_state = light_state

        # turn on requested, if we were off, use the dft_on_state as a base
        if state_changes["on_off"] == 1 and not light_state["on_off"]:
            _LOGGER.debug("Bulb was off, using dft_on_state")
            new_state = light_state["dft_on_state"]

        # override the existing settings
        new_state.update(state_changes)

        if (
            not state_changes["on_off"] and "dft_on_state" not in light_state
        ):  # if not already off, pack the data inside dft_on_state
            _LOGGER.debug(
                "Bulb was on and turn_off was requested, saving to dft_on_state"
            )
            new_state = {"dft_on_state": light_state, "on_off": 0}

        _LOGGER.debug("New light state: %s", new_state)
        self.proto["system"]["get_sysinfo"]["light_state"] = new_state

    def set_preferred_state(self, new_state, *args):
        """Implement set_preferred_state."""
        self.proto["system"]["get_sysinfo"]["preferred_state"][new_state["index"]] = (
            new_state
        )

    def light_state(self, x, *args):
        light_state = self.proto["system"]["get_sysinfo"]["light_state"]
        # Our tests have light state off, so we simply return the dft_on_state when device is on.
        _LOGGER.debug("reporting light state: %s", light_state)
        # TODO: hack to go around KL430 fixture differences
        if light_state["on_off"] and "dft_on_state" in light_state:
            return light_state["dft_on_state"]
        else:
            return light_state

    def set_time(self, new_state: dict, *args):
        """Implement set_time."""
        mods = [
            v
            for k, v in self.proto.items()
            if k in {"time", "smartlife.iot.common.timesetting"}
        ]
        index = new_state.pop("index", None)
        for mod in mods:
            mod["get_time"] = new_state
            if index is not None:
                mod["get_timezone"]["index"] = index

    baseproto = {
        "system": {
            "set_relay_state": set_relay_state,
            "set_dev_alias": set_alias,
            "set_led_off": set_led_off,
            "get_dev_icon": {"icon": None, "hash": None},
            "set_mac_addr": set_mac,
            "get_sysinfo": None,
        },
        "emeter": {
            "get_realtime": None,
            "get_daystat": None,
            "get_monthstat": None,
            "erase_emeter_state": None,
        },
        "smartlife.iot.common.emeter": {
            "get_realtime": None,
            "get_daystat": None,
            "get_monthstat": None,
            "erase_emeter_state": None,
        },
        "smartlife.iot.smartbulb.lightingservice": {
            "get_light_state": light_state,
            "get_light_details": LIGHT_DETAILS,
            "get_default_behavior": DEFAULT_BEHAVIOR,
            "transition_light_state": transition_light_state,
            "set_preferred_state": set_preferred_state,
        },
        "smartlife.iot.lighting_effect": {
            "set_lighting_effect": set_lighting_effect,
        },
        # lightstrip follows the same payloads but uses different module & method
        "smartlife.iot.lightStrip": {
            "set_light_state": transition_light_state,
            "get_light_state": light_state,
            "get_light_details": LIGHT_DETAILS,
            "get_default_behavior": DEFAULT_BEHAVIOR,
            "set_preferred_state": set_preferred_state,
        },
        "smartlife.iot.common.system": {
            "set_dev_alias": set_alias,
        },
        "time": {**TIME_MODULE, "set_time": set_time, "set_timezone": set_time},
        "smartlife.iot.common.timesetting": {
            **TIME_MODULE,
            "set_time": set_time,
            "set_timezone": set_time,
        },
        # HS220 brightness, different setter and getter
        "smartlife.iot.dimmer": {
            "set_brightness": set_hs220_brightness,
            "set_dimmer_transition": set_hs220_dimmer_transition,
        },
        "smartlife.iot.LAS": AMBIENT_MODULE,
        "smartlife.iot.PIR": MOTION_MODULE,
        "cnCloud": CLOUD_MODULE,
        "smartlife.iot.common.cloud": CLOUD_MODULE,
        "schedule": SCHEDULE_MODULE,
        "smartlife.iot.common.schedule": SCHEDULE_MODULE,
    }

    async def send(self, request, port=9999):
        if not self.verbatim:
            return await self._send(request, port)

        # Simply return whatever is in the fixture
        response = {}
        for target in request:
            if target in self.proto:
                response.update({target: self.proto[target]})
            else:
                response.update({"err_msg": "module not support"})
        return copy.deepcopy(response)

    async def _send(self, request, port=9999):
        proto = self.proto
        # collect child ids from context
        try:
            child_ids = request["context"]["child_ids"]
            request.pop("context", None)
        except KeyError:
            child_ids = []

        def get_response_for_module(target):
            if target not in proto:
                return error(msg="target not found")
            if "err_code" in proto[target] and proto[target]["err_code"] != 0:
                return {target: proto[target]}

            def get_response_for_command(cmd):
                if cmd not in proto[target]:
                    return error(msg=f"command {cmd} not found")

                params = request[target][cmd]
                _LOGGER.debug(
                    "Going to execute %s.%s (params: %s).. ", target, cmd, params
                )

                if callable(proto[target][cmd]):
                    res = proto[target][cmd](self, params, child_ids)
                    _LOGGER.debug("[callable] %s.%s: %s", target, cmd, res)
                    return success(res)
                elif isinstance(proto[target][cmd], dict):
                    res = proto[target][cmd]
                    _LOGGER.debug("[static] %s.%s: %s", target, cmd, res)
                    return success(res)
                else:
                    raise NotImplementedError(f"target {target} cmd {cmd}")

            from collections import defaultdict

            cmd_responses = defaultdict(dict)
            for cmd in request[target]:
                cmd_responses[target][cmd] = get_response_for_command(cmd)

            return cmd_responses

        response = {}
        for target in request:
            response.update(get_response_for_module(target))

        return copy.deepcopy(response)

    async def close(self) -> None:
        pass

    async def reset(self) -> None:
        pass



import logging
from unittest.mock import AsyncMock, patch

import pytest
from pytest_mock import MockerFixture

from kasa import Device, Feature, KasaException
from kasa.iot import IotStrip

_LOGGER = logging.getLogger(__name__)


class DummyDevice:
    pass


@pytest.fixture
def dummy_feature() -> Feature:
    # create_autospec for device slows tests way too much, so we use a dummy here

    feat = Feature(
        device=DummyDevice(),  # type: ignore[arg-type]
        id="dummy_feature",
        name="dummy_feature",
        attribute_getter="dummygetter",
        attribute_setter="dummysetter",
        container=None,
        icon="mdi:dummy",
        type=Feature.Type.Switch,
        unit_getter=lambda: "dummyunit",
    )
    return feat


def test_feature_api(dummy_feature: Feature):
    """Test all properties of a dummy feature."""
    assert dummy_feature.device is not None
    assert dummy_feature.name == "dummy_feature"
    assert dummy_feature.attribute_getter == "dummygetter"
    assert dummy_feature.attribute_setter == "dummysetter"
    assert dummy_feature.container is None
    assert dummy_feature.icon == "mdi:dummy"
    assert dummy_feature.type == Feature.Type.Switch
    assert dummy_feature.unit == "dummyunit"


@pytest.mark.parametrize(
    "read_only_type", [Feature.Type.Sensor, Feature.Type.BinarySensor]
)
def test_feature_setter_on_sensor(read_only_type):
    """Test that creating a sensor feature with a setter causes an error."""
    with pytest.raises(ValueError, match="Invalid type for configurable feature"):
        Feature(
            device=DummyDevice(),  # type: ignore[arg-type]
            id="dummy_error",
            name="dummy error",
            attribute_getter="dummygetter",
            attribute_setter="dummysetter",
            type=read_only_type,
        )


def test_feature_value(dummy_feature: Feature):
    """Verify that property gets accessed on *value* access."""
    dummy_feature.attribute_getter = "test_prop"
    dummy_feature.device.test_prop = "dummy"  # type: ignore[attr-defined]
    assert dummy_feature.value == "dummy"


def test_feature_value_container(mocker, dummy_feature: Feature):
    """Test that container's attribute is accessed when expected."""

    class DummyContainer:
        @property
        def test_prop(self):
            return "dummy"

    dummy_feature.container = DummyContainer()  # type: ignore[assignment]
    dummy_feature.attribute_getter = "test_prop"

    mock_dev_prop = mocker.patch.object(
        dummy_feature, "test_prop", new_callable=mocker.PropertyMock, create=True
    )

    assert dummy_feature.value == "dummy"
    mock_dev_prop.assert_not_called()


def test_feature_value_callable(dev, dummy_feature: Feature):
    """Verify that callables work as *attribute_getter*."""
    dummy_feature.attribute_getter = lambda x: "dummy value"
    assert dummy_feature.value == "dummy value"


async def test_feature_setter(dev, mocker, dummy_feature: Feature):
    """Verify that *set_value* calls the defined method."""
    mock_set_dummy = mocker.patch.object(
        dummy_feature.device, "set_dummy", create=True, new_callable=AsyncMock
    )
    dummy_feature.attribute_setter = "set_dummy"
    await dummy_feature.set_value("dummy value")
    mock_set_dummy.assert_called_with("dummy value")


async def test_feature_setter_read_only(dummy_feature):
    """Verify that read-only feature raises an exception when trying to change it."""
    dummy_feature.attribute_setter = None
    with pytest.raises(ValueError, match="Tried to set read-only feature"):
        await dummy_feature.set_value("value for read only feature")


async def test_feature_action(mocker):
    """Test that setting value on button calls the setter."""
    feat = Feature(
        device=DummyDevice(),  # type: ignore[arg-type]
        id="dummy_feature",
        name="dummy_feature",
        attribute_setter="call_action",
        container=None,
        icon="mdi:dummy",
        type=Feature.Type.Action,
    )
    mock_call_action = mocker.patch.object(
        feat.device, "call_action", create=True, new_callable=AsyncMock
    )
    assert feat.value == "<Action>"
    await feat.set_value(1234)
    mock_call_action.assert_called()


@pytest.mark.xdist_group(name="caplog")
async def test_feature_choice_list(dummy_feature, caplog, mocker: MockerFixture):
    """Test the choice feature type."""
    dummy_feature.type = Feature.Type.Choice
    dummy_feature.choices_getter = lambda: ["first", "second"]

    mock_setter = mocker.patch.object(
        dummy_feature.device, "dummysetter", create=True, new_callable=AsyncMock
    )
    await dummy_feature.set_value("first")
    mock_setter.assert_called_with("first")
    mock_setter.reset_mock()

    with pytest.raises(  # noqa: PT012
        ValueError,
        match="Unexpected value for dummy_feature: 'invalid' (?: - allowed: .*)?",
    ):
        await dummy_feature.set_value("invalid")
        assert "Unexpected value" in caplog.text

    mock_setter.assert_not_called()


@pytest.mark.parametrize("precision_hint", [1, 2, 3])
async def test_precision_hint(dummy_feature, precision_hint):
    """Test that precision hint works as expected."""
    dummy_value = 3.141593
    dummy_feature.type = Feature.Type.Sensor
    dummy_feature.precision_hint = precision_hint

    dummy_feature.attribute_getter = lambda x: dummy_value
    assert dummy_feature.value == dummy_value
    assert f"{round(dummy_value, precision_hint)} dummyunit" in repr(dummy_feature)


async def test_feature_setters(dev: Device, mocker: MockerFixture):
    """Test that all feature setters query something."""
    # setters that do not call set on the device itself.
    internal_setters = {"pan_step", "tilt_step"}

    async def _test_feature(feat, query_mock):
        if feat.attribute_setter is None:
            return

        # IotStrip makes calls via it's children
        expecting_call = feat.id not in internal_setters and not isinstance(
            dev, IotStrip
        )

        if feat.type == Feature.Type.Number:
            await feat.set_value(feat.minimum_value)
        elif feat.type == Feature.Type.Switch:
            await feat.set_value(True)
        elif feat.type == Feature.Type.Action:
            await feat.set_value("dummyvalue")
        elif feat.type == Feature.Type.Choice:
            await feat.set_value(feat.choices[0])
        elif feat.type == Feature.Type.Unknown:
            _LOGGER.warning("Feature '%s' has no type, cannot test the setter", feat)
            expecting_call = False
        else:
            raise NotImplementedError(f"set_value not implemented for {feat.type}")

        if expecting_call:
            query_mock.assert_called()

    async def _test_features(dev):
        exceptions = []
        for feat in dev.features.values():
            try:
                patch_dev = feat.container._device if feat.container else feat.device
                with (
                    patch.object(patch_dev.protocol, "query", name=feat.id) as query,
                    # patch update in case feature setter does an update
                    patch.object(patch_dev, "update"),
                ):
                    await _test_feature(feat, query)
            # we allow our own exceptions to avoid mocking valid responses
            except KasaException:
                pass
            except Exception as ex:
                ex.add_note(f"Exception when trying to set {feat} on {dev}")
                exceptions.append(ex)

        return exceptions

    exceptions = await _test_features(dev)

    for child in dev.children:
        exceptions.extend(await _test_features(child))

    if exceptions:
        raise ExceptionGroup(
            "Got exceptions while testing attribute_setters", exceptions
        )



"""Base implementation for SMART modules."""

from __future__ import annotations

import logging
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import TYPE_CHECKING, Any, Concatenate, ParamSpec, TypeVar

from ..exceptions import DeviceError, KasaException, SmartErrorCode
from ..module import Module

if TYPE_CHECKING:
    from .smartdevice import SmartDevice

_LOGGER = logging.getLogger(__name__)

_T = TypeVar("_T", bound="SmartModule")
_P = ParamSpec("_P")
_R = TypeVar("_R")


def allow_update_after(
    func: Callable[Concatenate[_T, _P], Coroutine[Any, Any, _R]],
) -> Callable[Concatenate[_T, _P], Coroutine[Any, Any, _R]]:
    """Define a wrapper to set _last_update_time to None.

    This will ensure that a module is updated in the next update cycle after
    a value has been changed.
    """

    @wraps(func)
    async def _async_wrap(self: _T, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        try:
            return await func(self, *args, **kwargs)
        finally:
            self._last_update_time = None

    return _async_wrap


def raise_if_update_error(func: Callable[[_T], _R]) -> Callable[[_T], _R]:
    """Define a wrapper to raise an error if the last module update was an error."""

    @wraps(func)
    def _wrap(self: _T) -> _R:
        if err := self._last_update_error:
            raise err
        return func(self)

    return _wrap


class SmartModule(Module):
    """Base class for SMART modules."""

    NAME: str
    #: Module is initialized, if the given component is available
    REQUIRED_COMPONENT: str | None = None
    #: Module is initialized, if any of the given keys exists in the sysinfo
    SYSINFO_LOOKUP_KEYS: list[str] = []
    #: Query to execute during the main update cycle
    QUERY_GETTER_NAME: str = ""

    REGISTERED_MODULES: dict[str, type[SmartModule]] = {}

    MINIMUM_UPDATE_INTERVAL_SECS = 0
    MINIMUM_HUB_CHILD_UPDATE_INTERVAL_SECS = 60 * 60 * 24

    UPDATE_INTERVAL_AFTER_ERROR_SECS = 30

    DISABLE_AFTER_ERROR_COUNT = 10

    def __init__(self, device: SmartDevice, module: str) -> None:
        self._device: SmartDevice
        super().__init__(device, module)
        self._last_update_time: float | None = None
        self._last_update_error: KasaException | None = None
        self._error_count = 0
        self._logged_remove_keys: list[str] = []

    def __init_subclass__(cls, **kwargs) -> None:
        # We only want to register submodules in a modules package so that
        # other classes can inherit from smartmodule and not be registered
        if cls.__module__.split(".")[-2] == "modules":
            _LOGGER.debug("Registering %s", cls)
            cls.REGISTERED_MODULES[cls._module_name()] = cls

    def _set_error(self, err: Exception | None) -> None:
        if err is None:
            self._error_count = 0
            self._last_update_error = None
        else:
            self._last_update_error = KasaException("Module update error", err)
            self._error_count += 1
            if self._error_count == self.DISABLE_AFTER_ERROR_COUNT:
                _LOGGER.error(
                    "Error processing %s for device %s, module will be disabled: %s",
                    self.name,
                    self._device.host,
                    err,
                )
            if self._error_count > self.DISABLE_AFTER_ERROR_COUNT:
                _LOGGER.error(  # pragma: no cover
                    "Unexpected error processing %s for device %s, "
                    "module should be disabled: %s",
                    self.name,
                    self._device.host,
                    err,
                )

    @property
    def update_interval(self) -> int:
        """Time to wait between updates."""
        if self._last_update_error:
            return self.UPDATE_INTERVAL_AFTER_ERROR_SECS * self._error_count

        if self._device._is_hub_child:
            return self.MINIMUM_HUB_CHILD_UPDATE_INTERVAL_SECS

        return self.MINIMUM_UPDATE_INTERVAL_SECS

    @property
    def disabled(self) -> bool:
        """Return true if the module is disabled due to errors."""
        return self._error_count >= self.DISABLE_AFTER_ERROR_COUNT

    def _should_update(self, update_time: float) -> bool:
        """Return true if module should update based on delay parameters."""
        return (
            not self.update_interval
            or not self._last_update_time
            or (update_time - self._last_update_time) >= self.update_interval
        )

    @classmethod
    def _module_name(cls) -> str:
        return getattr(cls, "NAME", cls.__name__)

    @property
    def name(self) -> str:
        """Name of the module."""
        return self._module_name()

    async def _post_update_hook(self) -> None:  # noqa: B027
        """Perform actions after a device update.

        Any modules overriding this should ensure that self.data is
        accessed unless the module should remain active despite errors.
        """
        assert self.data  # noqa: S101

    def query(self) -> dict:
        """Query to execute during the update cycle.

        Default implementation uses the raw query getter w/o parameters.
        """
        if self.QUERY_GETTER_NAME:
            return {self.QUERY_GETTER_NAME: None}
        return {}

    async def call(self, method: str, params: dict | None = None) -> dict:
        """Call a method.

        Just a helper method.
        """
        return await self._device._query_helper(method, params)

    @property
    def optional_response_keys(self) -> list[str]:
        """Return optional response keys for the module.

        Defaults to no keys. Overriding this and providing keys will remove
        instead of raise on error.
        """
        return []

    @property
    def data(self) -> dict[str, Any]:
        """Return response data for the module.

        If the module performs only a single query, the resulting response is unwrapped.
        If the module does not define a query, this property returns a reference
        to the main "get_device_info" response.
        """
        dev = self._device
        q = self.query()

        if not q:
            return dev.sys_info

        q_keys = list(q.keys())
        query_key = q_keys[0]

        # TODO: hacky way to check if update has been called.
        #  The way this falls back to parent may not always be wanted.
        #  Especially, devices can have their own firmware updates.
        if query_key not in dev._last_update:
            if dev._parent and query_key in dev._parent._last_update:
                _LOGGER.debug("%s not found child, but found on parent", query_key)
                dev = dev._parent
            else:
                raise KasaException(
                    f"You need to call update() prior accessing module data"
                    f" for '{self._module}'"
                )

        filtered_data = {k: v for k, v in dev._last_update.items() if k in q_keys}

        remove_keys: list[str] = []
        for data_item in filtered_data:
            if isinstance(filtered_data[data_item], SmartErrorCode):
                if data_item in self.optional_response_keys:
                    remove_keys.append(data_item)
                else:
                    raise DeviceError(
                        f"{data_item} for {self.name}",
                        error_code=filtered_data[data_item],
                    )

        for key in remove_keys:
            if key not in self._logged_remove_keys:
                self._logged_remove_keys.append(key)
                _LOGGER.debug(
                    "Removed key %s from response for device %s as it returned "
                    "error: %s. This message will only be logged once per key.",
                    key,
                    self._device.host,
                    filtered_data[key],
                )

            filtered_data.pop(key)

        if len(filtered_data) == 1 and not remove_keys:
            return next(iter(filtered_data.values()))

        return filtered_data

    @property
    def supported_version(self) -> int:
        """Return version supported by the device.

        If the module has no required component, this will return -1.
        """
        if self.REQUIRED_COMPONENT is not None:
            return self._device._components[self.REQUIRED_COMPONENT]
        return -1

    async def _check_supported(self) -> bool:
        """Additional check to see if the module is supported by the device.

        Used for parents who report components on the parent that are only available
        on the child or for modules where the device has a pointless component like
        color_temp_range but only supports one value.
        """
        return True

    def _has_data_error(self) -> bool:
        try:
            assert self.data  # noqa: S101
            return False
        except DeviceError:
            return True



"""Credentials class for username / passwords."""

from __future__ import annotations

import base64
from dataclasses import dataclass, field


@dataclass
class Credentials:
    """Credentials for authentication."""

    #: Username (email address) of the cloud account
    username: str = field(default="", repr=False)
    #: Password of the cloud account
    password: str = field(default="", repr=False)


def get_default_credentials(tuple: tuple[str, str]) -> Credentials:
    """Return decoded default credentials."""
    un = base64.b64decode(tuple[0].encode()).decode()
    pw = base64.b64decode(tuple[1].encode()).decode()
    return Credentials(un, pw)


DEFAULT_CREDENTIALS = {
    "KASA": ("a2FzYUB0cC1saW5rLm5ldA==", "a2FzYVNldHVw"),
    "KASACAMERA": ("YWRtaW4=", "MjEyMzJmMjk3YTU3YTVhNzQzODk0YTBlNGE4MDFmYzM="),
    "TAPO": ("dGVzdEB0cC1saW5rLm5ldA==", "dGVzdA=="),
    "TAPOCAMERA": ("YWRtaW4=", "YWRtaW4="),
}



"""Module for led controls."""

from __future__ import annotations

from ...interfaces.led import Led as LedInterface
from ..iotmodule import IotModule


class Led(IotModule, LedInterface):
    """Implementation of led controls."""

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {}

    @property
    def mode(self) -> str:
        """LED mode setting.

        "always", "never"
        """
        return "always" if self.led else "never"

    @property
    def led(self) -> bool:
        """Return the state of the led."""
        sys_info = self.data
        return bool(1 - sys_info["led_off"])

    async def set_led(self, state: bool) -> dict:
        """Set the state of the led (night mode)."""
        return await self.call("set_led_off", {"off": int(not state)})

    @property
    def is_supported(self) -> bool:
        """Return whether the module is supported by the device."""
        return "led_off" in self.data



"""Parse pcaps for TP-Link communications."""

import json
from collections import Counter, defaultdict
from pprint import pformat as pf

import click
import dpkt
from dpkt.ethernet import ETH_TYPE_IP, Ethernet

from kasa.cli.main import echo
from kasa.transports.xortransport import XorEncryption


def read_payloads_from_file(file):
    """Read the given pcap file and yield json payloads."""
    pcap = dpkt.pcap.Reader(file)
    for _ts, pkt in pcap:
        eth = Ethernet(pkt)
        if eth.type != ETH_TYPE_IP:
            continue

        ip = eth.ip
        if ip.p == 6:
            transport = ip.tcp
        elif ip == 17:
            transport = ip.udp
        else:
            continue

        if transport.sport != 9999 and transport.dport != 9999:
            continue

        data = transport.data

        try:
            decrypted = XorEncryption.decrypt(data[4:])
        except Exception as ex:
            echo(f"[red]Unable to decrypt the data, ignoring: {ex}[/red]")
            continue

        if not decrypted:  # skip empty payloads
            continue

        try:
            json_payload = json.loads(decrypted)
        except Exception as ex:
            # this can happen when the response is split into multiple tcp segments
            echo(f"[red]Unable to parse payload '{decrypted}', ignoring: {ex}[/red]")
            continue

        if not json_payload:  # ignore empty payloads
            echo("[red]Got empty payload, ignoring[/red]")
            continue

        yield json_payload


@click.command()
@click.argument("file", type=click.File("rb"))
def parse_pcap(file):
    """Parse pcap file and pretty print the communications and some statistics."""
    seen_items = defaultdict(Counter)

    for json_payload in read_payloads_from_file(file):
        context = json_payload.pop("context", "")
        for module, cmds in json_payload.items():
            seen_items["modules"][module] += 1
            if "err_code" in cmds:
                echo(f"[red]Got error for module: {cmds}[/red]")
                continue

            for cmd, response in cmds.items():
                seen_items["commands"][cmd] += 1
                seen_items["full_command"][f"{module}.{cmd}"] += 1
                if response is None:
                    continue
                direction = ">>"
                if response is None:
                    echo(f"got none as response for {cmd} %s, weird?")
                    continue
                is_success = "[green]+[/green]"
                if "err_code" in response:
                    direction = "<<"
                    if response["err_code"] != 0:
                        seen_items["errorcodes"][response["err_code"]] += 1
                        seen_items["errors"][response["err_msg"]] += 1
                        is_success = "[red]![/red]"

                context_str = f" [ctx: {context}]" if context else ""

                echo(
                    f"[{is_success}] {direction}{context_str} {module}.{cmd}:"
                    f" {pf(response)}"
                )

    echo(pf(seen_items))


if __name__ == "__main__":
    parse_pcap()



"""Module for SmartCamDevice."""

from __future__ import annotations

import logging
from typing import Any, cast

from ..device import DeviceInfo
from ..device_type import DeviceType
from ..module import Module
from ..protocols.smartcamprotocol import _ChildCameraProtocolWrapper
from ..smart import SmartChildDevice, SmartDevice
from ..smart.smartdevice import ComponentsRaw
from .modules import ChildDevice, DeviceModule
from .smartcammodule import SmartCamModule

_LOGGER = logging.getLogger(__name__)


class SmartCamDevice(SmartDevice):
    """Class for smart cameras."""

    # Modules that are called as part of the init procedure on first update
    FIRST_UPDATE_MODULES = {DeviceModule, ChildDevice}

    @staticmethod
    def _get_device_type_from_sysinfo(sysinfo: dict[str, Any]) -> DeviceType:
        """Find type to be displayed as a supported device category."""
        if not (device_type := sysinfo.get("device_type")):
            return DeviceType.Unknown

        if device_type.endswith("HUB"):
            return DeviceType.Hub

        if "DOORBELL" in device_type:
            return DeviceType.Doorbell

        return DeviceType.Camera

    @staticmethod
    def _get_device_info(
        info: dict[str, Any], discovery_info: dict[str, Any] | None
    ) -> DeviceInfo:
        """Get model information for a device."""
        basic_info = info["getDeviceInfo"]["device_info"]["basic_info"]
        short_name = basic_info["device_model"]
        long_name = discovery_info["device_model"] if discovery_info else short_name
        device_type = SmartCamDevice._get_device_type_from_sysinfo(basic_info)
        fw_version_full = basic_info["sw_version"]
        if " " in fw_version_full:
            firmware_version, firmware_build = fw_version_full.split(" ", maxsplit=1)
        else:
            firmware_version, firmware_build = fw_version_full, None
        return DeviceInfo(
            short_name=basic_info["device_model"],
            long_name=long_name,
            brand="tapo",
            device_family=basic_info["device_type"],
            device_type=device_type,
            hardware_version=basic_info["hw_version"],
            firmware_version=firmware_version,
            firmware_build=firmware_build,
            requires_auth=True,
            region=basic_info.get("region"),
        )

    def _update_internal_info(self, info_resp: dict) -> None:
        """Update the internal device info."""
        info = self._try_get_response(info_resp, "getDeviceInfo")
        self._info = self._map_info(info["device_info"])

    def _update_internal_state(self, info: dict[str, Any]) -> None:
        """Update the internal info state.

        This is used by the parent to push updates to its children.
        """
        self._info = self._map_info(info)

    async def _update_children_info(self) -> bool:
        """Update the internal child device info from the parent info.

        Return true if children added or deleted.
        """
        changed = False
        if child_info := self._try_get_response(
            self._last_update, "getChildDeviceList", {}
        ):
            changed = await self._create_delete_children(
                child_info, self._last_update["getChildDeviceComponentList"]
            )

            for info in child_info["child_device_list"]:
                child_id = info.get("device_id")
                if child_id not in self._children:
                    # _create_delete_children has already logged a message
                    continue

                self._children[child_id]._update_internal_state(info)

        return changed

    async def _initialize_smart_child(
        self, info: dict, child_components_raw: ComponentsRaw
    ) -> SmartDevice:
        """Initialize a smart child device attached to a smartcam device."""
        child_id = info["device_id"]
        child_protocol = _ChildCameraProtocolWrapper(child_id, self.protocol)
        try:
            initial_response = await child_protocol.query(
                {"get_connect_cloud_state": None}
            )
        except Exception as ex:
            _LOGGER.exception("Error initialising child %s: %s", child_id, ex)

        return await SmartChildDevice.create(
            parent=self,
            child_info=info,
            child_components_raw=child_components_raw,
            protocol=child_protocol,
            last_update=initial_response,
        )

    async def _initialize_smartcam_child(
        self, info: dict, child_components_raw: ComponentsRaw
    ) -> SmartDevice:
        """Initialize a smart child device attached to a smartcam device."""
        child_id = info["device_id"]
        child_protocol = _ChildCameraProtocolWrapper(child_id, self.protocol)

        app_component_list = {
            "app_component_list": child_components_raw["component_list"]
        }
        from .smartcamchild import SmartCamChild

        return await SmartCamChild.create(
            parent=self,
            child_info=info,
            child_components_raw=app_component_list,
            protocol=child_protocol,
        )

    async def _initialize_children(self) -> None:
        """Initialize children for hubs."""
        child_info_query = {
            "getChildDeviceList": {"childControl": {"start_index": 0}},
            "getChildDeviceComponentList": {"childControl": {"start_index": 0}},
        }
        resp = await self.protocol.query(child_info_query)
        self.internal_state.update(resp)

    async def _try_create_child(
        self, info: dict, child_components: dict
    ) -> SmartDevice | None:
        if not (category := info.get("category")):
            return None

        # Smart
        if category in SmartChildDevice.CHILD_DEVICE_TYPE_MAP:
            return await self._initialize_smart_child(info, child_components)
        # Smartcam
        from .smartcamchild import SmartCamChild

        if category in SmartCamChild.CHILD_DEVICE_TYPE_MAP:
            return await self._initialize_smartcam_child(info, child_components)

        return None

    async def _initialize_modules(self) -> None:
        """Initialize modules based on component negotiation response."""
        for mod in SmartCamModule.REGISTERED_MODULES.values():
            if (
                mod.REQUIRED_COMPONENT
                and mod.REQUIRED_COMPONENT not in self._components
            ):
                continue
            module = mod(self, mod._module_name())
            if await module._check_supported():
                self._modules[module.name] = module

    async def _initialize_features(self) -> None:
        """Initialize device features."""
        for module in self.modules.values():
            module._initialize_features()
            for feat in module._module_features.values():
                self._add_feature(feat)

    async def _query_setter_helper(
        self, method: str, module: str, section: str, params: dict | None = None
    ) -> dict:
        res = await self.protocol.query({method: {module: {section: params}}})

        return res

    @staticmethod
    def _parse_components(components_raw: ComponentsRaw) -> dict[str, int]:
        return {
            str(comp["name"]): int(comp["version"])
            for comp in components_raw["app_component_list"]
        }

    async def _negotiate(self) -> None:
        """Perform initialization.

        We fetch the device info and the available components as early as possible.
        If the device reports supporting child devices, they are also initialized.
        """
        initial_query = {
            "getDeviceInfo": {"device_info": {"name": ["basic_info", "info"]}},
            "getAppComponentList": {"app_component": {"name": "app_component_list"}},
            "getConnectionType": {"network": {"get_connection_type": {}}},
        }
        resp = await self.protocol.query(initial_query)
        self._last_update.update(resp)
        self._update_internal_info(resp)

        self._components_raw = cast(
            ComponentsRaw, resp["getAppComponentList"]["app_component"]
        )
        self._components = self._parse_components(self._components_raw)

        if "childControl" in self._components and not self.children:
            await self._initialize_children()

    def _map_info(self, device_info: dict) -> dict:
        """Map the basic keys to the keys used by SmartDevices."""
        basic_info = device_info["basic_info"]
        mappings = {
            "device_model": "model",
            "device_alias": "alias",
            "sw_version": "fw_ver",
            "hw_version": "hw_ver",
            "hw_id": "hwId",
            "dev_id": "device_id",
        }
        return {mappings.get(k, k): v for k, v in basic_info.items()}

    @property
    def is_on(self) -> bool:
        """Return true if the device is on."""
        if (camera := self.modules.get(Module.Camera)) and not camera.disabled:
            return camera.is_on

        return True

    async def set_state(self, on: bool) -> dict:
        """Set the device state."""
        if (camera := self.modules.get(Module.Camera)) and not camera.disabled:
            return await camera.set_state(on)

        return {}

    @property
    def device_type(self) -> DeviceType:
        """Return the device type."""
        if self._device_type == DeviceType.Unknown and self._info:
            self._device_type = self._get_device_type_from_sysinfo(self._info)
        return self._device_type

    @property
    def alias(self) -> str | None:
        """Returns the device alias or nickname."""
        if self._info:
            return self._info.get("alias")
        return None

    async def set_alias(self, alias: str) -> dict:
        """Set the device name (alias)."""
        return await self.protocol.query(
            {
                "setDeviceAlias": {"system": {"sys": {"dev_alias": alias}}},
            }
        )

    @property
    def hw_info(self) -> dict:
        """Return hardware info for the device."""
        return {
            "sw_ver": self._info.get("fw_ver"),
            "hw_ver": self._info.get("hw_ver"),
            "mac": self._info.get("mac"),
            "type": self._info.get("type"),
            "hwId": self._info.get("hwId"),
            "dev_name": self.alias,
            "oemId": self._info.get("oem_id"),
        }

    @property
    def rssi(self) -> int | None:
        """Return the device id."""
        return self.modules[SmartCamModule.SmartCamDeviceModule].rssi



"""Package containing all supported protocols."""

from .iotprotocol import IotProtocol
from .protocol import BaseProtocol
from .smartcamprotocol import SmartCamProtocol
from .smartprotocol import SmartErrorCode, SmartProtocol

__all__ = [
    "BaseProtocol",
    "IotProtocol",
    "SmartErrorCode",
    "SmartProtocol",
    "SmartCamProtocol",
]



"""Implementation of the dimmer config module found in dimmers."""

from __future__ import annotations

import logging
from datetime import timedelta
from typing import Any, Final, cast

from ...exceptions import KasaException
from ...feature import Feature
from ..iotmodule import IotModule, merge

_LOGGER = logging.getLogger(__name__)


def _td_to_ms(td: timedelta) -> int:
    """
    Convert timedelta to integer milliseconds.

    Uses default float to integer rounding.
    """
    return int(td / timedelta(milliseconds=1))


class Dimmer(IotModule):
    """Implements the dimmer config module."""

    THRESHOLD_ABS_MIN: Final[int] = 0
    # Strange value, but verified against hardware (KS220).
    THRESHOLD_ABS_MAX: Final[int] = 51
    FADE_TIME_ABS_MIN: Final[timedelta] = timedelta(seconds=0)
    # Arbitrary, but set low intending GENTLE FADE for longer fades.
    FADE_TIME_ABS_MAX: Final[timedelta] = timedelta(seconds=10)
    GENTLE_TIME_ABS_MIN: Final[timedelta] = timedelta(seconds=0)
    # Arbitrary, but reasonable default.
    GENTLE_TIME_ABS_MAX: Final[timedelta] = timedelta(seconds=120)
    # Verified against KS220.
    RAMP_RATE_ABS_MIN: Final[int] = 10
    # Verified against KS220.
    RAMP_RATE_ABS_MAX: Final[int] = 50

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="dimmer_threshold_min",
                name="Minimum dimming level",
                icon="mdi:lightbulb-on-20",
                attribute_getter="threshold_min",
                attribute_setter="set_threshold_min",
                range_getter=lambda: (self.THRESHOLD_ABS_MIN, self.THRESHOLD_ABS_MAX),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="dimmer_fade_off_time",
                name="Dimmer fade off time",
                icon="mdi:clock-in",
                attribute_getter="fade_off_time",
                attribute_setter="set_fade_off_time",
                range_getter=lambda: (
                    _td_to_ms(self.FADE_TIME_ABS_MIN),
                    _td_to_ms(self.FADE_TIME_ABS_MAX),
                ),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="dimmer_fade_on_time",
                name="Dimmer fade on time",
                icon="mdi:clock-out",
                attribute_getter="fade_on_time",
                attribute_setter="set_fade_on_time",
                range_getter=lambda: (
                    _td_to_ms(self.FADE_TIME_ABS_MIN),
                    _td_to_ms(self.FADE_TIME_ABS_MAX),
                ),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="dimmer_gentle_off_time",
                name="Dimmer gentle off time",
                icon="mdi:clock-in",
                attribute_getter="gentle_off_time",
                attribute_setter="set_gentle_off_time",
                range_getter=lambda: (
                    _td_to_ms(self.GENTLE_TIME_ABS_MIN),
                    _td_to_ms(self.GENTLE_TIME_ABS_MAX),
                ),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="dimmer_gentle_on_time",
                name="Dimmer gentle on time",
                icon="mdi:clock-out",
                attribute_getter="gentle_on_time",
                attribute_setter="set_gentle_on_time",
                range_getter=lambda: (
                    _td_to_ms(self.GENTLE_TIME_ABS_MIN),
                    _td_to_ms(self.GENTLE_TIME_ABS_MAX),
                ),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )

        self._add_feature(
            Feature(
                device=self._device,
                container=self,
                id="dimmer_ramp_rate",
                name="Dimmer ramp rate",
                icon="mdi:clock-fast",
                attribute_getter="ramp_rate",
                attribute_setter="set_ramp_rate",
                range_getter=lambda: (self.RAMP_RATE_ABS_MIN, self.RAMP_RATE_ABS_MAX),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )

    def query(self) -> dict:
        """Request Dimming configuration."""
        req = merge(
            self.query_for_command("get_dimmer_parameters"),
            self.query_for_command("get_default_behavior"),
        )

        return req

    @property
    def config(self) -> dict[str, Any]:
        """Return current configuration."""
        return self.data["get_dimmer_parameters"]

    @property
    def threshold_min(self) -> int:
        """Return the minimum dimming level for this dimmer."""
        return self.config["minThreshold"]

    async def set_threshold_min(self, min: int) -> dict:
        """Set the minimum dimming level for this dimmer.

        The value will depend on the luminaries connected to the dimmer.

        :param min: The minimum dimming level, in the range 0-51.
        """
        if min < self.THRESHOLD_ABS_MIN or min > self.THRESHOLD_ABS_MAX:
            raise KasaException(
                "Minimum dimming threshold is outside the supported range: "
                f"{self.THRESHOLD_ABS_MIN}-{self.THRESHOLD_ABS_MAX}"
            )
        return await self.call("calibrate_brightness", {"minThreshold": min})

    @property
    def fade_off_time(self) -> timedelta:
        """Return the fade off animation duration."""
        return timedelta(milliseconds=cast(int, self.config["fadeOffTime"]))

    async def set_fade_off_time(self, time: int | timedelta) -> dict:
        """Set the duration of the fade off animation.

        :param time: The animation duration, in ms.
        """
        if isinstance(time, int):
            time = timedelta(milliseconds=time)
        if time < self.FADE_TIME_ABS_MIN or time > self.FADE_TIME_ABS_MAX:
            raise KasaException(
                "Fade time is outside the bounds of the supported range:"
                f"{self.FADE_TIME_ABS_MIN}-{self.FADE_TIME_ABS_MAX}"
            )
        return await self.call("set_fade_off_time", {"fadeTime": _td_to_ms(time)})

    @property
    def fade_on_time(self) -> timedelta:
        """Return the fade on animation duration."""
        return timedelta(milliseconds=cast(int, self.config["fadeOnTime"]))

    async def set_fade_on_time(self, time: int | timedelta) -> dict:
        """Set the duration of the fade on animation.

        :param time: The animation duration, in ms.
        """
        if isinstance(time, int):
            time = timedelta(milliseconds=time)
        if time < self.FADE_TIME_ABS_MIN or time > self.FADE_TIME_ABS_MAX:
            raise KasaException(
                "Fade time is outside the bounds of the supported range:"
                f"{self.FADE_TIME_ABS_MIN}-{self.FADE_TIME_ABS_MAX}"
            )
        return await self.call("set_fade_on_time", {"fadeTime": _td_to_ms(time)})

    @property
    def gentle_off_time(self) -> timedelta:
        """Return the gentle fade off animation duration."""
        return timedelta(milliseconds=cast(int, self.config["gentleOffTime"]))

    async def set_gentle_off_time(self, time: int | timedelta) -> dict:
        """Set the duration of the gentle fade off animation.

        :param time: The animation duration, in ms.
        """
        if isinstance(time, int):
            time = timedelta(milliseconds=time)
        if time < self.GENTLE_TIME_ABS_MIN or time > self.GENTLE_TIME_ABS_MAX:
            raise KasaException(
                "Gentle off time is outside the bounds of the supported range: "
                f"{self.GENTLE_TIME_ABS_MIN}-{self.GENTLE_TIME_ABS_MAX}."
            )
        return await self.call("set_gentle_off_time", {"duration": _td_to_ms(time)})

    @property
    def gentle_on_time(self) -> timedelta:
        """Return the gentle fade on animation duration."""
        return timedelta(milliseconds=cast(int, self.config["gentleOnTime"]))

    async def set_gentle_on_time(self, time: int | timedelta) -> dict:
        """Set the duration of the gentle fade on animation.

        :param time: The animation duration, in ms.
        """
        if isinstance(time, int):
            time = timedelta(milliseconds=time)
        if time < self.GENTLE_TIME_ABS_MIN or time > self.GENTLE_TIME_ABS_MAX:
            raise KasaException(
                "Gentle off time is outside the bounds of the supported range: "
                f"{self.GENTLE_TIME_ABS_MIN}-{self.GENTLE_TIME_ABS_MAX}."
            )
        return await self.call("set_gentle_on_time", {"duration": _td_to_ms(time)})

    @property
    def ramp_rate(self) -> int:
        """Return the rate that the dimmer buttons increment the dimmer level."""
        return self.config["rampRate"]

    async def set_ramp_rate(self, rate: int) -> dict:
        """Set how quickly to ramp the dimming level when using the dimmer buttons.

        :param rate: The rate to increment the dimming level with each press.
        """
        if rate < self.RAMP_RATE_ABS_MIN or rate > self.RAMP_RATE_ABS_MAX:
            raise KasaException(
                "Gentle off time is outside the bounds of the supported range:"
                f"{self.RAMP_RATE_ABS_MIN}-{self.RAMP_RATE_ABS_MAX}"
            )
        return await self.call("set_button_ramp_rate", {"rampRate": rate})



from __future__ import annotations

import asyncio
import copy
from collections.abc import Coroutine
from dataclasses import dataclass
from json import dumps as json_dumps
from typing import Any, TypedDict

import pytest

from kasa.transports.xortransport import XorEncryption

from .fakeprotocol_iot import FakeIotProtocol
from .fakeprotocol_smart import FakeSmartProtocol, FakeSmartTransport
from .fakeprotocol_smartcam import FakeSmartCamProtocol
from .fixtureinfo import FixtureInfo, filter_fixtures, idgenerator

DISCOVERY_MOCK_IP = "127.0.0.123"


class DiscoveryResponse(TypedDict):
    result: dict[str, Any]
    error_code: int


UNSUPPORTED_HOMEWIFISYSTEM = {
    "error_code": 0,
    "result": {
        "channel_2g": "10",
        "channel_5g": "44",
        "device_id": "REDACTED_51f72a752213a6c45203530",
        "device_model": "X20",
        "device_type": "HOMEWIFISYSTEM",
        "factory_default": False,
        "group_id": "REDACTED_07d902da02fa9beab8a64",
        "group_name": "I01BU0tFRF9TU0lEIw==",  # '#MASKED_SSID#'
        "hardware_version": "3.0",
        "ip": "127.0.0.1",
        "mac": "24:2F:D0:00:00:00",
        "master_device_id": "REDACTED_51f72a752213a6c45203530",
        "need_account_digest": True,
        "owner": "REDACTED_341c020d7e8bda184e56a90",
        "role": "master",
        "tmp_port": [20001],
    },
}


def _make_unsupported(
    device_family,
    encrypt_type,
    *,
    https: bool = False,
    omit_keys: dict[str, Any] | None = None,
) -> DiscoveryResponse:
    if omit_keys is None:
        omit_keys = {"encrypt_info": None}
    result: DiscoveryResponse = {
        "result": {
            "device_id": "xx",
            "owner": "xx",
            "device_type": device_family,
            "device_model": "P110(EU)",
            "ip": "127.0.0.1",
            "mac": "48-22xxx",
            "is_support_iot_cloud": True,
            "obd_src": "tplink",
            "factory_default": False,
            "mgt_encrypt_schm": {
                "is_support_https": https,
                "encrypt_type": encrypt_type,
                "http_port": 80,
                "lv": 2,
            },
            "encrypt_info": {"data": "", "key": "", "sym_schm": encrypt_type},
        },
        "error_code": 0,
    }
    for key, val in omit_keys.items():
        if val is None:
            result["result"].pop(key)
        else:
            result["result"][key].pop(val)

    return result


UNSUPPORTED_DEVICES = {
    "unknown_device_family": _make_unsupported("SMART.TAPOXMASTREE", "AES"),
    "unknown_iot_device_family": _make_unsupported("IOT.IOTXMASTREE", "AES"),
    "wrong_encryption_iot": _make_unsupported("IOT.SMARTPLUGSWITCH", "AES"),
    "wrong_encryption_smart": _make_unsupported("SMART.TAPOBULB", "IOT"),
    "unknown_encryption": _make_unsupported("IOT.SMARTPLUGSWITCH", "FOO"),
    "missing_encrypt_type": _make_unsupported(
        "SMART.TAPOBULB",
        "FOO",
        omit_keys={"mgt_encrypt_schm": "encrypt_type", "encrypt_info": None},
    ),
    "unable_to_parse": _make_unsupported(
        "SMART.TAPOBULB",
        "FOO",
        omit_keys={"device_id": None},
    ),
    "invalidinstance": _make_unsupported(
        "IOT.SMARTPLUGSWITCH",
        "KLAP",
        https=True,
    ),
    "homewifi": UNSUPPORTED_HOMEWIFISYSTEM,
}


def parametrize_discovery(
    desc, *, data_root_filter=None, protocol_filter=None, model_filter=None
):
    filtered_fixtures = filter_fixtures(
        desc,
        data_root_filter=data_root_filter,
        protocol_filter=protocol_filter,
        model_filter=model_filter,
    )
    return pytest.mark.parametrize(
        "discovery_mock",
        filtered_fixtures,
        indirect=True,
        ids=idgenerator,
    )


new_discovery = parametrize_discovery(
    "new discovery", data_root_filter="discovery_result"
)

smart_discovery = parametrize_discovery("smart discovery", protocol_filter={"SMART"})


@pytest.fixture(
    params=filter_fixtures(
        "discoverable", protocol_filter={"SMART", "SMARTCAM", "IOT"}
    ),
    ids=idgenerator,
)
async def discovery_mock(request, mocker):
    """Mock discovery and patch protocol queries to use Fake protocols."""
    fi: FixtureInfo = request.param
    fixture_info = FixtureInfo(fi.name, fi.protocol, copy.deepcopy(fi.data))
    return patch_discovery({DISCOVERY_MOCK_IP: fixture_info}, mocker)


def create_discovery_mock(ip: str, fixture_data: dict):
    """Mock discovery and patch protocol queries to use Fake protocols."""

    @dataclass
    class _DiscoveryMock:
        ip: str
        default_port: int
        discovery_port: int
        discovery_data: dict
        query_data: dict
        device_type: str
        encrypt_type: str
        https: bool
        login_version: int | None = None
        port_override: int | None = None
        http_port: int | None = None

        @property
        def model(self) -> str:
            dd = self.discovery_data
            model_region = (
                dd["result"]["device_model"]
                if self.discovery_port == 20002
                else dd["system"]["get_sysinfo"]["model"]
            )
            model, _, _ = model_region.partition("(")
            return model

        @property
        def _datagram(self) -> bytes:
            if self.default_port == 9999:
                return XorEncryption.encrypt(json_dumps(self.discovery_data))[4:]
            else:
                return (
                    b"\x02\x00\x00\x01\x01[\x00\x00\x00\x00\x00\x00W\xcev\xf8"
                    + json_dumps(self.discovery_data).encode()
                )

    if "discovery_result" in fixture_data:
        discovery_data = fixture_data["discovery_result"].copy()
        discovery_result = fixture_data["discovery_result"]["result"]
        device_type = discovery_result["device_type"]
        encrypt_type = discovery_result["mgt_encrypt_schm"].get(
            "encrypt_type", discovery_result.get("encrypt_info", {}).get("sym_schm")
        )

        if not (login_version := discovery_result["mgt_encrypt_schm"].get("lv")) and (
            et := discovery_result.get("encrypt_type")
        ):
            login_version = max([int(i) for i in et])
        https = discovery_result["mgt_encrypt_schm"]["is_support_https"]
        http_port = discovery_result["mgt_encrypt_schm"].get("http_port")
        if not http_port:  # noqa: SIM108
            # Not all discovery responses set the http port, i.e. smartcam.
            default_port = 443 if https else 80
        else:
            default_port = http_port
        dm = _DiscoveryMock(
            ip,
            default_port,
            20002,
            discovery_data,
            fixture_data,
            device_type,
            encrypt_type,
            https,
            login_version,
            http_port=http_port,
        )
    else:
        sys_info = fixture_data["system"]["get_sysinfo"]
        discovery_data = {"system": {"get_sysinfo": sys_info.copy()}}
        device_type = sys_info.get("mic_type") or sys_info.get("type")
        encrypt_type = "XOR"
        login_version = None
        dm = _DiscoveryMock(
            ip,
            9999,
            9999,
            discovery_data,
            fixture_data,
            device_type,
            encrypt_type,
            False,
            login_version,
        )

    return dm


def patch_discovery(fixture_infos: dict[str, FixtureInfo], mocker):
    """Mock discovery and patch protocol queries to use Fake protocols."""
    discovery_mocks = {
        ip: create_discovery_mock(ip, fixture_info.data)
        for ip, fixture_info in fixture_infos.items()
    }
    protos = {
        ip: FakeSmartProtocol(fixture_info.data, fixture_info.name)
        if fixture_info.protocol in {"SMART", "SMART.CHILD"}
        else FakeSmartCamProtocol(fixture_info.data, fixture_info.name)
        if fixture_info.protocol in {"SMARTCAM", "SMARTCAM.CHILD"}
        else FakeIotProtocol(fixture_info.data, fixture_info.name)
        for ip, fixture_info in fixture_infos.items()
    }
    first_ip = list(fixture_infos.keys())[0]
    first_host = None

    # Mock _run_callback_task so the tasks complete in the order they started.
    # Otherwise test output is non-deterministic which affects readme examples.
    callback_queue: asyncio.Queue = asyncio.Queue()
    exception_queue: asyncio.Queue = asyncio.Queue()

    async def process_callback_queue(finished_event: asyncio.Event) -> None:
        while (finished_event.is_set() is False) or callback_queue.qsize():
            coro = await callback_queue.get()
            try:
                await coro
            except Exception as ex:
                await exception_queue.put(ex)
            else:
                await exception_queue.put(None)
            callback_queue.task_done()

    async def wait_for_coro():
        await callback_queue.join()
        if ex := exception_queue.get_nowait():
            raise ex

    def _run_callback_task(self, coro: Coroutine) -> None:
        callback_queue.put_nowait(coro)
        task = asyncio.create_task(wait_for_coro())
        self.callback_tasks.append(task)

    mocker.patch(
        "kasa.discover._DiscoverProtocol._run_callback_task", _run_callback_task
    )

    # do_discover_mock
    async def mock_discover(self):
        """Call datagram_received for all mock fixtures.

        Handles test cases modifying the ip and hostname of the first fixture
        for discover_single testing.
        """
        finished_event = asyncio.Event()
        asyncio.create_task(process_callback_queue(finished_event))

        for ip, dm in discovery_mocks.items():
            first_ip = list(discovery_mocks.values())[0].ip
            fixture_info = fixture_infos[ip]
            # Ip of first fixture could have been modified by a test
            if dm.ip == first_ip:
                # hostname could have been used
                host = first_host if first_host else first_ip
            else:
                host = dm.ip
            # update the protos for any host testing or the test overriding the first ip
            protos[host] = (
                FakeSmartProtocol(fixture_info.data, fixture_info.name)
                if fixture_info.protocol in {"SMART", "SMART.CHILD"}
                else FakeSmartCamProtocol(fixture_info.data, fixture_info.name)
                if fixture_info.protocol in {"SMARTCAM", "SMARTCAM.CHILD"}
                else FakeIotProtocol(fixture_info.data, fixture_info.name)
            )
            port = (
                dm.port_override
                if dm.port_override and dm.discovery_port != 20002
                else dm.discovery_port
            )
            self.datagram_received(
                dm._datagram,
                (dm.ip, port),
            )
        # Setting this event will stop the processing of callbacks
        finished_event.set()

    mocker.patch("kasa.discover._DiscoverProtocol.do_discover", mock_discover)

    # query_mock
    async def _query(self, request, retry_count: int = 3):
        return await protos[self._host].query(request)

    mocker.patch("kasa.IotProtocol.query", _query)
    mocker.patch("kasa.SmartProtocol.query", _query)

    def _getaddrinfo(host, *_, **__):
        nonlocal first_host, first_ip
        first_host = host  # Store the hostname used by discover single
        first_ip = list(discovery_mocks.values())[
            0
        ].ip  # ip could have been overridden in test
        return [(None, None, None, None, (first_ip, 0))]

    mocker.patch("socket.getaddrinfo", side_effect=_getaddrinfo)

    # Mock decrypt so it doesn't error with unencryptable empty data in the
    # fixtures. The discovery result will already contain the decrypted data
    # deserialized from the fixture
    mocker.patch("kasa.discover.Discover._decrypt_discovery_data")

    # Only return the first discovery mock to be used for testing discover single
    return discovery_mocks[first_ip]


@pytest.fixture(
    params=filter_fixtures(
        "discoverable", protocol_filter={"SMART", "SMARTCAM", "IOT"}
    ),
    ids=idgenerator,
)
def discovery_data(request, mocker):
    """Return raw discovery file contents as JSON. Used for discovery tests."""
    fixture_info = request.param
    fixture_data = copy.deepcopy(fixture_info.data)
    # Add missing queries to fixture data
    if "component_nego" in fixture_data:
        components = {
            comp["id"]: int(comp["ver_code"])
            for comp in fixture_data["component_nego"]["component_list"]
        }
        for k, v in FakeSmartTransport.FIXTURE_MISSING_MAP.items():
            # Value is a tuple of component,reponse
            if k not in fixture_data and v[0] in components:
                fixture_data[k] = v[1]
    mocker.patch("kasa.IotProtocol.query", return_value=fixture_data)
    mocker.patch("kasa.SmartProtocol.query", return_value=fixture_data)
    if "discovery_result" in fixture_data:
        return fixture_data["discovery_result"].copy()
    else:
        return {"system": {"get_sysinfo": fixture_data["system"]["get_sysinfo"]}}


@pytest.fixture(
    params=UNSUPPORTED_DEVICES.values(), ids=list(UNSUPPORTED_DEVICES.keys())
)
def unsupported_device_info(request, mocker):
    """Return unsupported devices for cli and discovery tests."""
    discovery_data = request.param
    host = "127.0.0.1"

    async def mock_discover(self):
        if discovery_data:
            data = (
                b"\x02\x00\x00\x01\x01[\x00\x00\x00\x00\x00\x00W\xcev\xf8"
                + json_dumps(discovery_data).encode()
            )
            self.datagram_received(data, (host, 20002))

    mocker.patch("kasa.discover._DiscoverProtocol.do_discover", mock_discover)

    return discovery_data
